/**************************************************************************/
/*                                                                        */
/*  Thibaut Balabonski, Sylvain Conchon, Jean-Christophe Filliâtre,       */
/*  Kim Nguyen, Laurent Sartre                                            */
/*                                                                        */
/*  Informatique - MP2I/MPI - CPGE 1re et 2e années.                      */
/*  Cours et exercices corrigés. Éditions Ellipses, 2022.                 */
/*                                                                        */
/*  https://www.informatique-mpi.fr/                                      */
/*                                                                        */
/**************************************************************************/

#include "prelude.h"

int binom_rec(int k, int n) {
  assert (0 <= k && k <= n);
  if (k == 0 || k == n)
    return 1;
  else
    return binom_rec(k-1, n-1) + binom_rec(k, n-1);
}

int binom_memo_aux(int k, int n, int **triangle) {
  if (k == 0 || k == n)
    return 1;
  else {
    if (triangle[n][k] == 0)
      triangle[n][k] = binom_memo_aux(k-1, n-1, triangle) + binom_memo_aux(k, n-1, triangle);
    return triangle[n][k];
  }
}
/* int binom_memo_aux(int k, int n, int **triangle) { */
/*   if (k == 0 || k == n) */
/*     return 1; */
/*   else if (triangle[n][k] != 0) */
/*     return triangle[n][k]; */
/*   else { */
/*     return triangle[n][k] = binom_memo_aux(k-1, n-1, triangle) + binom_memo_aux(k, n-1, triangle); */
/*   } */
/* } */
int binom_memo(int k, int n) {
  int **triangle = calloc(n + 1, sizeof(int*));
  for (int i = 0; i <= n; i++) {
    triangle[i] = calloc(i + 1, sizeof(int));
  }
  return binom_memo_aux(k, n, triangle);
}

int binom_full_triangle(int k, int n) {
  int **triangle = calloc(n + 1, sizeof(int*));
  for (int i = 0; i <= n; i++) {
    triangle[i] = calloc(i + 1, sizeof(int));
    triangle[i][0] = 1;
    triangle[i][i] = 1;
    for (int j = 1; j < i; j++) {
      triangle[i][j] = triangle[i-1][j-1] + triangle[i-1][j];
    }
  }
  return triangle[n][k];
}

int binom_two_rows(int k, int n) {
  int *row1 = calloc(n + 1, sizeof(int));
  int *row2 = calloc(n + 1, sizeof(int));
  int *rowtmp;
  for (int i = 0; i <= n; i++) {
    row2[0] = 1;
    row2[i] = 1;
    for (int j = 1; j < i; j++) {
      row2[j] = row1[j-1] + row1[j];
    }
    rowtmp = row1;
    row1 = row2;
    row2 = rowtmp;
  }
  return row1[k];
}

int binom_single_row(int k, int n) {
  int *row = calloc(n + 1, sizeof(int));
  for (int i = 0; i <= n; i++) {
    row[i] = 1;
    for (int j = i - 1; j > 0; j--)
      row[j] += row[j-1];
  }
  return row[k];
}

int main() {
  int k = 4, n = 7;
  int br = binom_rec(k, n);
  int bm = binom_memo(k, n);
  int bf = binom_full_triangle(k, n);
  int bt = binom_two_rows(k, n);
  int bs = binom_single_row(k, n);
  printf("%d parmi %d : rec %d, memo %d, full %d, tr %d, sr %d\n",
         k, n, br, bm, bf, bt, bs);
}
